//----------------------------------------------------------------------------
// $Id: Reader.java,v 1.31 2004/05/10 22:49:07 enz Exp $
// $Source: /cvsroot/gogui/gogui/src/sgf/Reader.java,v $
//----------------------------------------------------------------------------

package sgf;

//----------------------------------------------------------------------------

import java.io.*;
import java.util.*;

import version.Version;
import game.*;
import go.*;
import gtp.Gtp;
import gui.HGui;

//----------------------------------------------------------------------------

public class Reader
{
    public static class Error extends Exception
    {
        public Error(String s)
        {
            super(s);
        }
    }

    boolean isLittleGolemGame=false;

    /** @param name Name prepended to error messages */
    public Reader(java.io.Reader reader, String name)
        throws Error
    {
        try
        {
            m_in = new BufferedReader(reader);
            m_tokenizer = new StreamTokenizer(m_in);
            m_name = name;
            m_tokenizer.nextToken();
            if (m_tokenizer.ttype != '(')
                throw getError("No root tree found.");
            Node root = new Node();
            Node node = readNext(root, true);
            while (node != null)
                node = readNext(node, false);
            if (root.getNumberChildren() == 1)
            {
                root = root.getChild();
                root.setFather(null);
            }
            checkTree(root, -1, false); //and apply fix for "swap-sides"
            m_gameTree = new GameTree(m_gameInformation, root);
            applyFixes();
        }
        catch (FileNotFoundException e)
        {
            throw new Error("File not found.");
        }
        catch (IOException e)
        {
            throw new Error("Error while reading file.");
        }
    }

    public GameTree getGameTree()
    {
        return m_gameTree;
    }

    private java.io.Reader m_in;

    private GameInformation m_gameInformation = new GameInformation(Version.defaultBoardSize);

    private GameTree m_gameTree;

    private StreamTokenizer m_tokenizer;

    private String m_name;

    // checkMovesInTree
    // and check for B / W move sequences
    // int sideSwapped is actually  (0 or 1)
    private void checkTree(Node root, int moveNr, boolean sideSwapped) throws Error {
      Move move = root.getMove();
      if (moveNr==-1) {
        if (move!=null) throw new Error("Error: Root Node has a move.");
      } else {
        if (move.getColor()!=Color.WHITE && move.getColor()!=Color.BLACK) {
          throw new Error("Error: Color must be BLACK OR WHITE!");
        }
        if (sideSwapped) { // APPLY swap-sides CORRECTION to the node
          move.setColor(move.getColor().otherColor());
          move.swapCoords();
        }
        Color col_expected=(moveNr%2==0) ? Color.BLACK : Color.WHITE;
        if (move==null) throw new Error("Error: non-root Node without move!");
        if (move.getColor() != col_expected)
          throw new Error("Error: unexpected Color in Node: Move("+move.toString(true)+"), at depth="+moveNr+" / expected: "+col_expected.toString().toUpperCase()+" found: "+move.getColor().toString().toUpperCase()+" (need: alternating Black / White)");
        if (moveNr==1 && move.getSpecialMove()==Move.MOVE_SWAP_SIDES) {
          move.setSpecialMove(Move.MOVE_SWAP);
          sideSwapped=true;
        }
      }
      for (int i = 0; i < root.getNumberChildren(); ++i) {
        Node n = root.getChild(i);
        checkTree(n, moveNr+1, sideSwapped);
      }
    }

    /** Apply some fixes for broken SGF files. */
    private void applyFixes()
    {
        Node root = m_gameTree.getRoot();
        GameInformation gameInformation = m_gameTree.getGameInformation();
        if ((root.getNumberAddWhite() + root.getNumberAddBlack() > 0)
            && root.getPlayer() == Color.EMPTY)
        {
            if (gameInformation.m_handicap > 0)
            {
                root.setPlayer(Color.WHITE);
            }
            else
            {
                boolean hasBlackChildMoves = false;
                boolean hasWhiteChildMoves = false;
                for (int i = 0; i < root.getNumberChildren(); ++i)
                {
                    Move move = root.getChild(i).getMove();
                    if (move == null)
                        continue;
                    if (move.getColor() == Color.BLACK)
                        hasBlackChildMoves = true;
                    if (move.getColor() == Color.WHITE)
                        hasWhiteChildMoves = true;
                }
                if (hasBlackChildMoves && ! hasWhiteChildMoves)
                    root.setPlayer(Color.BLACK);
                if (hasWhiteChildMoves && ! hasBlackChildMoves)
                    root.setPlayer(Color.WHITE);
            }
        }
    }

    private Error getError(String message)
    {
        int lineNumber = m_tokenizer.lineno() + 1;
        if (m_name != null)
        {
            String s = m_name + ":" + lineNumber + ": " + message;
            return new Error(s);
        }
        else
            return new Error(lineNumber + ": " + message);
    }

    private Color parseColor(String s) throws Error
    {
        Color c;
        s = s.trim().toLowerCase();
        if (s.equals("b") || s.equals("1"))
            c = Color.BLACK;
        else if (s.equals("w") || s.equals("2"))
            c = Color.WHITE;
        else
            throw getError("Invalid color value.");
        return c;
    }

    private float parseFloat(String s) throws Error
    {
        float f = 0;
        try
        {
            f = Float.parseFloat(s);
        }
        catch (NumberFormatException e)
        {
            throw getError("Floating point number expected.");
        }
        return f;
    }

    private int parseInt(String s) throws Error
    {
        int i = -1;
        try
        {
            i = Integer.parseInt(s);
        }
        catch (NumberFormatException e)
        {
            throw getError("Number expected.");
        }
        return i;
    }

    private int numCharactersAtBegin(String s) {
        int i=0;
        while (i<s.length()) {
            if (Character.isLetter(s.charAt(i))) i++; else break;
        }
        return i;
    }

    // NEW:
    public Move parseMove(Color c, String s) throws Error
    {
        try {
            return Move.parseMove(c, s, m_gameInformation.m_boardSize, true);
        }
        catch (Exception e) {
            throw new Error("Reader: "+e.getMessage());
        }
    }

    private Point parsePoint(String s) throws Error
    {
        s = s.trim().toLowerCase();
        if (s.equals(""))
            return null;
        if (s.length() < 2)
            throw getError("Invalid coordinates: " + s);
        int boardSize = m_gameInformation.m_boardSize;
        // ***************************
        int x,y;
        x = s.charAt(0) - 'a';
        int num=numCharactersAtBegin(s);
        if ((s.length()==2)&&(num==2)) {
            y = s.charAt(1) - 'a';
        } else { // a1..z19
            String s1=s.substring(num);
            try { y=Integer.parseInt(s1, 10)-1; } catch(NumberFormatException e) {
                throw getError("Invalid coordinates: " + s);
            }
        }
        // **********************************************
        if (x < 0 || x >= boardSize || y < 0 || y >= boardSize) //#
            throw getError("Invalid coordinates: " + s);
        return new Point(x, y);
    }

    private Node readNext(Node father, boolean isRoot)
        throws IOException, Error
    {
        m_tokenizer.nextToken();
        int ttype = m_tokenizer.ttype;
        if (ttype == '(')
        {
            Node node = father;
            while (node != null)
                node = readNext(node, false);
            return father;
        }
        if (ttype == ')')
            return null;
        if (ttype != ';')
            throw getError("Next node expected");
        Node son = new Node();
        father.append(son);
        while (readProp(son, isRoot));
        return son;
    }

    private boolean readProp(Node node, boolean isRoot)
        throws IOException, Error
    {
        m_tokenizer.nextToken();
        int ttype = m_tokenizer.ttype;
        if (ttype == StreamTokenizer.TT_WORD)
        {
            String p = m_tokenizer.sval.toUpperCase();
            Vector values = new Vector();
            String s;
            while ((s = readValue()) != null)
                values.add(s);
            if (values.size() == 0)
                throw getError("Property '" + p + "' has no value.");
            String v = (String)values.get(0);
            if (p.equals("AE"))
            {
                throw getError("Add empty not supported.");
            }
            else if (p.equals("AB") || p.equals("AW")) {
                throw getError("Add Black/White not supported.");
            }
            else if (p.equals("B") || p.equals("W")) {
              boolean addBlack=p.equals("B");
              if (isLittleGolemGame) addBlack=!addBlack;
              if (addBlack)
                  node.setMove(parseMove(Color.BLACK, v));
              else
                  node.setMove(parseMove(Color.WHITE, v));
            }
            else if (p.equals("BL"))
            {
                try
                {
                    node.setTimeLeftBlack(Float.parseFloat(v));
                }
                catch (NumberFormatException e)
                {
                }
            }
            else if (p.equals("BR"))
                m_gameInformation.m_blackRank = v;
            else if (p.equals("C"))
                node.setComment(v.trim());
            else if (p.equals("DT"))
                m_gameInformation.m_date = v;
            else if (p.equals("GM"))
            {
                v = v.trim();
                // Value should be 1, but sgf2misc produces Go files
                // with empty value
                if (!v.equals("Hex") && !v.equals("") && !v.equals("11"))
                    throw getError("Not a Hex game.");
            }
            else if (p.equals("HA"))
                m_gameInformation.m_handicap = Integer.parseInt(v);
            else if (p.equals("KM"))
                m_gameInformation.m_komi = parseFloat(v);
            else if (p.equals("OB"))
            {
                try
                {
                    node.setMovesLeftBlack(Integer.parseInt(v));
                }
                catch (NumberFormatException e)
                {
                }
            }
            else if (p.equals("OW"))
            {
                try
                {
                    node.setMovesLeftWhite(Integer.parseInt(v));
                }
                catch (NumberFormatException e)
                {
                }
            }
            else if (p.equals("PB")) {
                m_gameInformation.m_playerBlack = v;
            }
            else if (p.equals("PW")) {
                m_gameInformation.m_playerWhite = v;
            }
            else if (p.equals("PL"))
                node.setPlayer(parseColor(v));
            else if (p.equals("RE"))
                m_gameInformation.m_result = v;
            else if (p.equals("RU"))
                m_gameInformation.m_rules = v;
            else if (isRoot && p.equals("SO") && (-1!=v.indexOf("littlegolem"))) {
                isLittleGolemGame=true;
                System.out.println("Detected little golem game => exchanging B[] and W[]!!");
                node.setComment(v);
            }
            else if (p.equals("SZ"))
            {
                if (! isRoot)
                    throw getError("Size property outside root node.");
                m_gameInformation.m_boardSize = parseInt(v);
            }
            else if (p.equals("WL"))
            {
                try
                {
                    node.setTimeLeftWhite(Float.parseFloat(v));
                }
                catch (NumberFormatException e)
                {
                }
            }
            else if (p.equals("WR"))
                m_gameInformation.m_whiteRank = v;
            else if (p.equals("AP"))
            {
                //ignore
            }
            else if (! p.equals("FF") && ! p.equals("GN") && ! p.equals("AP")) {
                node.addSgfProperty(p, v);
            }
            return true;
        }
        m_tokenizer.pushBack();
        return false;
    }

    // read escaped prop value, including "[" and "]"
    private String readValue() throws IOException, Error
    {
        m_tokenizer.nextToken();
        int ttype = m_tokenizer.ttype;
        if (ttype != '[')
        {
            m_tokenizer.pushBack();
            return null;
        }
        String v = "";
        boolean quoted = false;
        while (true)
        {
            int c = m_in.read();
            if (c < 0)
                throw getError("Property value incomplete.");

            if (!quoted) {
                     if (c == ']') break; //FINISHED!
                else if (c == '\\') quoted=true;
                else v += (char)c; // add single character
            }
            else {
                     if (c=='t')  v += '\t';
                else if (c=='r')  v += '\r';
                else if (c=='n')  v += '\n';
                else if (c=='\r') continue; //quoted newline may (should) follow!! (thus => keep quoted=true)
                else if (c=='\n') ; //this is a quoted newline => ignore!
                else              v += (char)c; // add charcter after quote character as it is

                quoted=false;
            }
        }
        return v;
    }
}

//----------------------------------------------------------------------------
